import os
import re
import pandas as pd

def ezclean(microbiome_data, metadata, level):
    # Check the file extension for microbiome_data
    microbiome_ext = os.path.splitext(microbiome_data)[1].lower().lstrip('.')
    if microbiome_ext not in ['csv', 'xls', 'xlsx']:
        return "The file is not csv, xls, or xlsx format. Please check the file type for microbiome data."
    
    ##code checks the microbiome file upto this point
    
    ###     NOW THE FORMAT OF METADATA
    # Check the file extension for metadata
    metadata_ext = os.path.splitext(metadata)[1].lower().lstrip('.')
    if metadata_ext == "txt":
        metadata_df = pd.read_csv(metadata, sep="\t", header=0)
    elif metadata_ext == "csv":
        metadata_df = pd.read_csv(metadata, sep=",", header=0)
    elif metadata_ext in ["xls", "xlsx"]:
        metadata_df = pd.read_excel(metadata, header=0)
    else:
        return "Please check the file format of metadata."
    
    
    ####METADATA HEADER MATCHING
    # Checking the first header of metadata
    valid_headers = ["id", "sampleid", "sample id", "sample-id", "featureid", "feature id", "feature-id"]
    # Ensure that the header is trimmed of any leading or trailing whitespace and converted to lower case
    if metadata_df.columns[0].strip().lower() not in valid_headers:
        return "Please check the first header of the metadata for file format correction."




#####NOW READING THE FILE MICROBIOME FILE

    if microbiome_ext == "csv":
        microbiome_df = pd.read_csv(microbiome_data)
    elif microbiome_ext in ["xls", "xlsx"]:
        microbiome_df = pd.read_excel(microbiome_data, skiprows=1)  # Skipping index row


# Define the levels map with all variations
    levels_map = {
        "domain": "d__", "Domain": "d__", "DOMAIN": "d__", "D": "d__", "d": "d__",
        "kingdom": "d__", "Kingdom": "d__", "KINGDOM": "d__", "K": "d__", "k": "d__",
        "phylum": "p__", "Phylum": "p__", "PHYLUM": "p__", "P": "p__", "p": "p__",
        "class": "c__", "Class": "c__", "CLASS": "c__", "C": "c__", "c": "c__",
        "order": "o__", "Order": "o__", "ORDER": "o__", "O": "o__", "o": "o__",
        "family": "f__", "Family": "f__", "FAMILY": "f__", "F": "f__", "f": "f__",
        "genera": "g__", "Genera": "g__", "GENERA": "g__", "G": "g__", "g": "g__",
        "species": "s__", "Species": "s__", "SPECIES": "s__", "S": "s__", "s": "s__"
    }

    # Convert input level parameter to the correct taxonomy prefix
    level_key = levels_map.get(level.lower())
    if level_key is None:
        return "The level value should be one of the following: domain, phylum, class, order, family, genera, species or their abbreviations."
    level_value = level_key

#UPTO THIS POINT I DID THE FILE HANDLING AND THE LEVEL HANDLING WITH APPROPRIIATE ERROR MESSAGE 
#NOW THE FILES ARE INTO THE FUNCTION AND THE LEVEL IS READY LETS DO  THE REAL WORK
#####################----------------------------------------------------------------------------------------------------------------#####################

    ####separate THE MICROBIOMRE ONLY AND THE METADATA ONLY
    # Remove columns in microbiome data that are not in metadata headers (common columns)
    common_cols = list(set(microbiome_df.columns).intersection(set(metadata_df.columns)))
    just_microbiome_df = microbiome_df.loc[:, ~microbiome_df.columns.isin(common_cols)]
    just_metadata_df = metadata_df.loc[:, metadata_df.columns.isin(common_cols)]
    
    # Save the files
    just_microbiome_df.to_excel("just_microbiome.xlsx", index=False)
    just_metadata_df.to_excel("just_metadata.xlsx", index=False)





#####################--------------------------------------------------------------

#transposing the just microbiome
##Real transposing


    # Read the entire Excel file without assigning any row as header.
    orig = pd.read_excel("just_microbiome.xlsx", header=None)
    
    # Define the new column headers from the original first column (skip the top-left cell)
    new_columns = orig.iloc[1:, 0].tolist()
    
    # Define the new row index from the original header row (skip the top-left cell)
    new_index = orig.iloc[0, 1:].tolist()
    
    # Extract the data block that excludes the original header row and first column.
    data_block = orig.iloc[1:, 1:]
    
    # Transpose the data block so that:
    # - The rows become the new row labels (matching new_index)
    # - The columns become the new column headers (matching new_columns)
    data_transposed = data_block.T
    
    # Create the new DataFrame with the desired orientation.
    df_new = pd.DataFrame(data_transposed.values, index=new_index, columns=new_columns)
    
    # Save the new DataFrame to an Excel file.
    df_new.to_excel("microbiome_ezy_2.xlsx", index=True)


    #NOTE FOR MYSELF  SINCE I USE THE MEZY 1 WHICH WAS IN TEXT AND MEZY2 WAS THE ONE WITH NUMERICAL
    #I DID THE DIRECT NUMERICAL SO I SKIPPED MEZY_1 AND NAMED IT AS MEZY_2
    
    
    
    
    
    
#####################--------------------------------------------------------------   
    
    #ADDING TAXA COLUMN
    
    
        # ---- Additional Blank Column Block Start ----
    # Read the file "microbiome_ezy_2.xlsx" (all values as saved) without column names
    df_ezy2 = pd.read_excel("microbiome_ezy_2.xlsx", header=None, dtype=str)
    
    # Convert the data frame to a "matrix" by working with DataFrame slices
    # Extract the first column (as a DataFrame)
    first_col = df_ezy2.iloc[:, [0]]
    # Create a new DataFrame column with a blank value for each row
    blank_col = pd.DataFrame([""] * df_ezy2.shape[0])
    # Extract the remaining columns (all except the first)
    remaining_cols = df_ezy2.iloc[:, 1:]
    
    # Combine the first column, the blank column, and the remaining columns
    # This mimics cbind(m_ezy2[, 1, drop = FALSE], rep("", nrow(m_ezy2)), m_ezy2[, -1, drop = FALSE])
    new_matrix = pd.concat([first_col, blank_col, remaining_cols], axis=1, ignore_index=True)
    
    # Set the header (first row, second column) to "Taxa"
    new_matrix.iat[0, 1] = "Taxa"
    
    # Write the new matrix as "microbiome_ezy_3.xlsx" without row or column names
    new_matrix.to_excel("microbiome_ezy_3.xlsx", index=False, header=False)
    # ---- Additional Blank Column Block End ----



#####################--------------------------------------------------------------   

#REPLACING F__ WITH @ FOR ALL THE LEVELS

    # ---- Additional Level Replacement Block Start ----
    # Read the "microbiome_ezy_3.xlsx" file (all values as saved) without column names
    df_ezy3 = pd.read_excel("microbiome_ezy_3.xlsx", header=None, dtype=str)
    
    # Ensure the first column is treated as text
    df_ezy3[0] = df_ezy3[0].astype(str)
    
    # Replace the level_value (e.g., "f__" for family) with "@" in the first column
    df_ezy3[0] = df_ezy3[0].str.replace(level_value, "@", regex=False)
    
    # Save the updated data as "microbiome_ezy_4.xlsx" without row names and without auto-generated column names
    df_ezy3.to_excel("microbiome_ezy_4.xlsx", index=False, header=False)
    # ---- Additional Level Replacement Block End ----


#####################--------------------------------------------------------------   



#####EXTRACTING ANYTHING BETWEEN @ AND  ; AS THE TAXA NAME



    
    # ---- Additional Extraction and Paste Block Start ----
    # Read the "microbiome_ezy_4.xlsx" file (all data, including header) without column names
    df_ezy4 = pd.read_excel("microbiome_ezy_4.xlsx", header=None, dtype=str)
    
    # Process only if there is more than just the header row
    if df_ezy4.shape[0] > 1:
        # For each data row (excluding the header), extract the text between "@" and ";" from the first column
        def extract_text(cell):
            if "@" not in cell:
                return ""
            match = re.search(r".*@([^;]*);.*", cell)
            return match.group(1).strip() if match else ""
        
        # Apply extraction on all rows except the header (row 0) for the first column (column 0)
        extracted_text = df_ezy4.iloc[1:, 0].apply(extract_text)
        
        # Paste the extracted text into the second column for data rows (the header remains unchanged)
        df_ezy4.iloc[1:, 1] = extracted_text
    
    # Save the updated data as "microbiome_ezy_5.xlsx" without extra row or column names
    df_ezy4.to_excel("microbiome_ezy_5.xlsx", index=False, header=False)
    # ---- Additional Extraction and Paste Block End ----


####REMOVING THE FIRST COLUMNS

    # ---- Additional Delete First Column Block Start ----
    # Read the "microbiome_ezy_5.xlsx" file (all data, including header) without column names
    df_ezy5 = pd.read_excel("microbiome_ezy_5.xlsx", header=None, dtype=str)
    
    # Delete the first column entirely
    df_ezy5 = df_ezy5.drop(columns=0)
    
    # Save the updated data as "microbiome_ezy_6.xlsx" without row names and column headers
    df_ezy5.to_excel("microbiome_ezy_6.xlsx", index=False, header=False)
    # ---- Additional Delete First Column Block End ----


####CONVERTING TO THE NUMERICALS


    # ---- Additional Numeric Conversion Block for ezy_7 Start ----
    # Read the "microbiome_ezy_6.xlsx" file (all values as text) without column names
    df_ezy6 = pd.read_excel("microbiome_ezy_6.xlsx", header=None, dtype=str)
    
    # Separate header row (first row) and data rows (remaining rows)
    header_row_6 = df_ezy6.iloc[0, :].tolist()
    data_rows_6 = df_ezy6.iloc[1:, :].copy()
    
    # Convert all columns except the first in data_rows_6 to numeric
    for col in data_rows_6.columns[1:]:
        data_rows_6[col] = pd.to_numeric(data_rows_6[col], errors='coerce')
    
    # Create a new Excel file and write the header row as text (transposed) and data rows starting at row 2
    with pd.ExcelWriter("microbiome_ezy_7.xlsx", engine="openpyxl") as writer:
        # Write the header row as a single-row DataFrame at row 1
        pd.DataFrame([header_row_6]).to_excel(writer, sheet_name="Sheet1", index=False, header=False, startrow=0)
        # Write the data rows starting at row 2 (i.e., startrow=1)
        data_rows_6.to_excel(writer, sheet_name="Sheet1", index=False, header=False, startrow=1)
    # ---- Additional Numeric Conversion Block for ezy_7 End ----


 
########## HANDLING THE SAME TAXA NAME AND THE BLANKS

# ---- Additional Aggregation Block Start ----
    # Read the "microbiome_ezy_7.xlsx" file (all data, including the header row) without column names
    df_ezy7 = pd.read_excel("microbiome_ezy_7.xlsx", header=None)
    
    # Separate the header row (first row) and the data rows (remaining rows)
    header_row_7 = df_ezy7.iloc[0, :].tolist()
    data_rows_7 = df_ezy7.iloc[1:, :].copy()
    
    # Ensure the first column is treated as text and all other columns are numeric
    data_rows_7[0] = data_rows_7[0].astype(str)
    for col in data_rows_7.columns[1:]:
        data_rows_7[col] = pd.to_numeric(data_rows_7[col], errors='coerce')
    
    # Define replacement names based on the level_value
    other_names = {
        "d__": "Other_domains",
        "p__": "Other_phyla",
        "c__": "Other_classes",
        "o__": "Other_orders",
        "f__": "Other_families",
        "g__": "Other_genera",
        "s__": "Other_species"
    }
    
    # Determine the appropriate replacement name using the level_value.
    replacement_name = other_names.get(level_value, "Other")
    
    # Replace rows with missing taxa information (NA, blank, "#VALUE", or "nan" as a string)
    data_rows_7[0] = data_rows_7[0].apply(
        lambda x: replacement_name if (pd.isna(x) or str(x).strip() in ["", "#VALUE", "nan"]) else x
    )
    
    # Aggregate the data by the taxa (first) column:
    # For rows with the exact same taxa value, sum the numeric columns column-wise.
    aggregated_df = data_rows_7.groupby(0, as_index=False).sum()
    
    # Set the column names of the aggregated data using the original header row.
    new_header = header_row_7.copy()
    new_header[0] = "Taxa"
    aggregated_df.columns = new_header
    
    # Write the aggregated data (including header) to "microbiome_ezy_8.xlsx"
    aggregated_df.to_excel("microbiome_ezy_8.xlsx", index=False)
    # ---- Additional Aggregation Block End ----


#####################--------------------------------------------------------------   


###3333 NOW CONVERTING INTO THE PERCENTAGE

    
    # ---- Additional Column-wise Percentage Block Start ----
    # Read the "microbiome_ezy_8.xlsx" file with header information
    microbiome_ezy8 = pd.read_excel("microbiome_ezy_8.xlsx", header=0)
    
    # Extract the taxa column (first column) and numeric columns (columns 2 onward)
    taxa_column = microbiome_ezy8.iloc[:, 0]
    numeric_columns = microbiome_ezy8.iloc[:, 1:].copy()
    
    # Ensure all numeric columns are properly converted to numeric
    numeric_columns = numeric_columns.apply(pd.to_numeric, errors='coerce')
    
    # Calculate the column-wise percentages:
    # For each numeric column, divide each element by the column sum (ignoring NAs) and multiply by 100.
    # Note: Use axis=1 to align the sums with the columns.
    col_sums = numeric_columns.sum(axis=0, skipna=True)
    percentage_columns = numeric_columns.div(col_sums, axis=1) * 100
    
    # Combine the taxa column with the calculated percentage columns
    microbiome_ezy9 = pd.concat([taxa_column, percentage_columns], axis=1)
    
    # Save the resulting data frame as "microbiome_ezy_9.xlsx" without row names
    microbiome_ezy9.to_excel("microbiome_ezy_9.xlsx", index=False)
    # ---- Additional Column-wise Percentage Block End ----



#####################--------------------------------------------------------------   


###TRANSPOSE THE EXCEL



    # Read the entire Excel file (microbiome_ezy_9.xlsx) without assigning any row as a header.
    df_ezy_9 = pd.read_excel("microbiome_ezy_9.xlsx", header=None)
    
    # Define new column headers from the original first column (skipping the top-left cell)
    col_headers_ezy_9 = df_ezy_9.iloc[1:, 0].tolist()
    
    # Define new row labels from the original header row (skipping the top-left cell)
    row_labels_ezy_9 = df_ezy_9.iloc[0, 1:].tolist()
    
    # Extract the data block that excludes the original header row and first column.
    data_slice_ezy_9 = df_ezy_9.iloc[1:, 1:]
    
    # Transpose the data block so that:
    # - The rows become the new row labels (matching row_labels_ezy_9)
    # - The columns become the new column headers (matching col_headers_ezy_9)
    data_trans_ezy_9 = data_slice_ezy_9.T
    
    # Create the new DataFrame with the desired orientation.
    df_ezy_10 = pd.DataFrame(data_trans_ezy_9.values, index=row_labels_ezy_9, columns=col_headers_ezy_9)
    
    # Save the new DataFrame to an Excel file (microbiome_ezy_10.xlsx).
    df_ezy_10.to_excel("microbiome_ezy_10.xlsx", index=True)




####33CHANGE THE FIRST HEADER FROM THE METADATA 


    # ---- Additional Header Replacement Block Start ----
    # Extract the header of the first column from the metadata file
    metadata_first_header = metadata_df.columns[0]
    
    # Read the "microbiome_ezy_10.xlsx" file; since it was written without column names,
    # the first row actually contains the header information.
    df_ezy10 = pd.read_excel("microbiome_ezy_10.xlsx", header=None, dtype=str)
    
    # Extract the first row as the current header and convert to a list of strings
    new_header = df_ezy10.iloc[0, :].astype(str).tolist()
    
    # Replace the first element with the metadata header
    new_header[0] = metadata_first_header
    
    # Remove the first row (which contained the original header) from the data
    df_ezy10_data = df_ezy10.iloc[1:, :].copy()
    
    # Assign the new header to the data
    df_ezy10_data.columns = new_header
    
    # Save the updated data as "microbiome_ezy_11.xlsx" with proper column names
    df_ezy10_data.to_excel("microbiome_ezy_11.xlsx", index=False, header=True)
    # ---- Additional Header Replacement Block End ----




#####EVERYTHING IS  TEXT FOR THE EASE OF DATA HANDLING SO I WILL CONVERT THE NUMERICALS





    # ---- Additional Numeric Conversion Block for ezy_12 Start ----
    # Read the "microbiome_ezy_11.xlsx" file (all values as text) without column names
    df_ezy_11 = pd.read_excel("microbiome_ezy_11.xlsx", header=None, dtype=str)
    
    # Separate header row (first row) and data rows (remaining rows)
    header_row_ezy_11 = df_ezy_11.iloc[0, :].tolist()
    data_rows_ezy_11 = df_ezy_11.iloc[1:, :].copy()
    
    # Convert all columns except the first in data_rows_ezy_11 to numeric
    for col in data_rows_ezy_11.columns[1:]:
        data_rows_ezy_11[col] = pd.to_numeric(data_rows_ezy_11[col], errors='coerce')
    
    # Create a new Excel file and write the header row as text (transposed) and data rows starting at row 2
    with pd.ExcelWriter("microbiome_ezy_12.xlsx", engine="openpyxl") as writer:
        # Write the header row as a single-row DataFrame at row 1
        pd.DataFrame([header_row_ezy_11]).to_excel(writer, sheet_name="Sheet1", index=False, header=False, startrow=0)
        # Write the data rows starting at row 2 (i.e., startrow=1)
        data_rows_ezy_11.to_excel(writer, sheet_name="Sheet1", index=False, header=False, startrow=1)
    
    
    
#####333 NOW I WILL ADD THE METADATA COLUMNS TO THE CLEANED DATA SET



    # ---- Additional Metadata Merge Block Start ----
    # Read the "microbiome_ezy_12.xlsx" file (with header)
    df_ezy12 = pd.read_excel("microbiome_ezy_12.xlsx", header=0)
    
    # In the metadata, the first column is the key.
    # Instead of using meta_remaining (which lacks the key column),
    # build a keyed DataFrame from the full metadata.
    metadata_keyed = metadata_df.set_index(metadata_df.columns[0])
    
    # Extract the key values from df_ezy12 (the cleaned data)
    df_key = df_ezy12.iloc[:, 0]
    
    # Reindex the keyed metadata with df_key.
    # Then select only the columns corresponding to the original meta_remaining,
    # i.e. all columns except the first.
    appended_metadata = metadata_keyed.reindex(df_key)[metadata_df.columns[1:]].reset_index(drop=True)
    
    # Combine the df_ezy12 data with the appended metadata columns.
    # Reset the index of df_ezy12 to ensure proper alignment.
    df_ezy13 = pd.concat([df_ezy12.reset_index(drop=True), appended_metadata], axis=1)
    
    # Save the combined data as "microbiome_ezy_13.xlsx" with headers
    df_ezy13.to_excel("microbiome_ezy_13.xlsx", index=False)
    # ---- Additional Metadata Merge Block End ----
    
    



##### REMOVING ANY INTERMEDIATE FILES IN THE WORKING DIR 


    # List of intermediate files to be removed (excluding the final file "microbiome_ezy_13.xlsx")
    intermediate_files = [
        "just_microbiome.xlsx",
        "just_metadata.xlsx",
        "microbiome_ezy_2.xlsx",
        "microbiome_ezy_3.xlsx",
        "microbiome_ezy_4.xlsx",
        "microbiome_ezy_5.xlsx",
        "microbiome_ezy_6.xlsx",
        "microbiome_ezy_7.xlsx",
        "microbiome_ezy_8.xlsx",
        "microbiome_ezy_9.xlsx",
        "microbiome_ezy_10.xlsx",
        "microbiome_ezy_11.xlsx",
        "microbiome_ezy_12.xlsx"
    ]
    
    # Remove each intermediate file if it exists
    for f in intermediate_files:
        if os.path.exists(f):
            os.remove(f)

##FINAL RETURN BASED ON THE TAXA LEVEL IN THE INPUT OF THE FUNCTION

    # ---- Final Return Block ----
    final_names = {
        "d__": "mbX_cleaned_domains_or_kingdom.xlsx",
        "p__": "mbX_cleaned_phylum.xlsx",
        "c__": "mbX_cleaned_classes.xlsx",
        "o__": "mbX_cleaned_orders.xlsx",
        "f__": "mbX_cleaned_families.xlsx",
        "g__": "mbX_cleaned_genera.xlsx",
        "s__": "mbX_cleaned_species.xlsx"
    }
    
    final_file_name = final_names[level_value]
    os.rename("microbiome_ezy_13.xlsx", final_file_name)
    
    return final_file_name

